/*
 *  Copyright (C) 1998  Luca Deri <deri@unipi.it>
 *                      Stefano Suin <stefano@unipi.it>
 *                      
 * 			Centro SERRA, University of Pisa
 * 			http://www-serra.unipi.it/
 *  					
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * Copyright (c) 1994, 1996
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/* #define DEBUG */

#include "ntop.h"

#include <sys/file.h>
#include <sys/stat.h>

#if __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include <unistd.h>

#include "interface.h"

/* Extern */
extern unsigned int localnet, netmask;
extern char* intoa(struct in_addr addr);
extern struct timeval lastPktTime;
extern FILE *logd;
extern int32_t thisZone;

/* Forward */
int32_t gmt2local(time_t t);

/* Local */
#define MAX_NUM_NETWORKS      32
static short     numLocalNets=0;
static u_int32_t networks[MAX_NUM_NETWORKS][2]; /* [0]=network, [1]=mask */

/* VARARGS */
void
#if __STDC__
error(const char *fmt, ...)
#else
error(fmt, va_alist)
	const char *fmt;
	va_dcl
#endif
{
	va_list ap;

	(void)fprintf(stderr, "%s: ", program_name);
#if __STDC__
	va_start(ap, fmt);
#else
	va_start(ap);
#endif
	(void)vfprintf(stderr, fmt, ap);
	va_end(ap);
	if (*fmt) {
		fmt += strlen(fmt);
		if (fmt[-1] != '\n')
			(void)fputc('\n', stderr);
	}
	exit(1);
	/* NOTREACHED */
}

/*
 * Copy arg vector into a new buffer, concatenating arguments with spaces.
 */
char *
copy_argv(register char **argv)
{
	register char **p;
	register u_int len = 0;
	char *buf;
	char *src, *dst;

	p = argv;
	if (*p == 0)
		return 0;

	while (*p)
		len += strlen(*p++) + 1;

	buf = (char *)malloc(len);
	if (buf == NULL)
		error("copy_argv: malloc");

	p = argv;
	dst = buf;
	while ((src = *p++) != NULL) {
		while ((*dst++ = *src++) != '\0')
			;
		dst[-1] = ' ';
	}
	dst[-1] = '\0';

	return buf;
}

/* ********************************* */

short isBroadcastAddress(struct in_addr *addr) {
  if((addr->s_addr == 0x0)
     || ((addr->s_addr | netmask) == 0xffffffff)
     || ((addr->s_addr & 0x000000ff) == 0x000000ff)
     || ((addr->s_addr & 0x000000ff) == 0x00000000) /* Network address */
     )
    return 1;
  else
    return 0;
}

/* ********************************* */

short isLocalAddress(struct in_addr *addr) {
  if((addr->s_addr & netmask) == localnet) {
#ifdef DEBUG
     printf("%s is local\n", intoa(*addr));
#endif
    return 1;
  } else {
#ifdef DEBUG
    printf("%s is remote\n", intoa(*addr)); 
#endif
    /* Broadcast is considered a local address */
    return (isBroadcastAddress(addr));
  }
}

/* ********************************* */

/* Example: "131.114.0.0/16,193.43.104.0/24" */
void handleLocalAddresses(char* addresses) {
  char *address = strtok(addresses, ",");

  while(address != NULL) {
    char *mask = strchr(address, '/');
    
    if(mask == NULL)
      printf("Unknown network '%s'. It has been ignored.\n", address);
    else {
      char tmpStr[255];
      u_int32_t network;      
      int bits=atoi(&mask[1]), a, b, c, d;

      mask[0] = '\0';

      sscanf(address, "%d.%d.%d.%d", &a, &b, &c, &d);
#ifdef DEBUG
      printf("%d.%d.%d.%d - %d\n", a, b, c, d, bits);
#endif
      ((unsigned char*)&network)[0] = (unsigned char)a;
      ((unsigned char*)&network)[1] = (unsigned char)b;
      ((unsigned char*)&network)[2] = (unsigned char)c;
      ((unsigned char*)&network)[3] = (unsigned char)d;

      netmask = 0xffffffff >> bits;
      netmask = ~netmask;

#ifdef DEBUG
      printf("%d.%d.%d.%d - %d [0x%08x/0x%08x]\n", a, b, c, d, bits, network, netmask);
#endif
      if(numLocalNets < MAX_NUM_NETWORKS) {
	networks[numLocalNets][0] = network; /* network */
	networks[numLocalNets][1] = netmask; /* netmask */
	numLocalNets++;
      } else
	printf("Unable to handle network (too many entries!).\n");
    }

    address = strtok(NULL, ",");
  }
}

/* ********************************* */

/* This function returns true when a host is considered local
   as specified using the 'm' flag */
short isPseudoLocalAddress(struct in_addr *addr) {
  if((addr->s_addr & netmask) == localnet) {
    return 1;
  } else {
    int i;

    for(i=0; i<numLocalNets; i++) {
#ifdef DEBUG
      printf("%s [0x%08x/0x%08x]\n", 
	     intoa(*addr),
	     (addr->s_addr & networks[i][1] /* netmask */),
	     networks[i][0] /* network */);
#endif
      if((addr->s_addr & networks[i][1] /* netmask */) == networks[i][0] /* network */) {
	return 1;
      }
    }

    /* Broadcast is considered a local address */
    return(isBroadcastAddress(addr));
  }
}

/* ********************************* */

/*
 * Print the log timestamp
 */
void printLogTime()
{
  /* Unix timeval style */
  register int s;
  
  s = (lastPktTime.tv_sec + thisZone) % 86400;
  fprintf(logd, "%02d:%02d:%02d.%06u ",
	  s/3600, (s%3600)/60, s%60,
	  (u_int32_t)lastPktTime.tv_usec);
}

/* ********************************* */

/*
 * Returns the difference between gmt and local time in seconds.
 * Use gmtime() and localtime() to keep things simple.
 * [Borrowed from tcpdump]
 */
int32_t gmt2local(time_t t)
{
  register int dt, dir;
  register struct tm *gmt, *loc;
  struct tm sgmt;

  if (t == 0)
    t = time(NULL);
  gmt = &sgmt;
  *gmt = *gmtime(&t);
  loc = localtime(&t);
  dt = (loc->tm_hour - gmt->tm_hour) * 60 * 60 +
    (loc->tm_min - gmt->tm_min) * 60;

  /*
   * If the year or julian day is different, we span 00:00 GMT
   * and must add or subtract a day. Check the year first to
   * avoid problems when the julian day wraps.
   */
  dir = loc->tm_year - gmt->tm_year;
  if (dir == 0)
    dir = loc->tm_yday - gmt->tm_yday;
  dt += dir * 24 * 60 * 60;

  return (dt);
}
